use std::{
    ffi::c_void,
    net::{SocketAddr, TcpListener},
    os::fd::IntoRawFd,
    sync::LazyLock,
    thread,
};

use network::{LogicScrollMessageFactory, Messaging, RC4Encrypter};
use session::PlayerSession;
use tracing::info;

mod allocator;
mod array_list;
mod ffi_util;
mod logic;
mod network;
mod resources;
mod sc_string;
mod session;
mod util;

#[allow(non_snake_case)]
#[unsafe(no_mangle)]
pub unsafe extern "system" fn JNI_OnLoad(_vm: jni::JavaVM, _: *mut c_void) -> jni::sys::jint {
    init_tracing();

    thread::spawn(|| server_main());

    jni::sys::JNI_VERSION_1_6
}

fn server_main() {
    const TCP_ADDR: &str = "127.0.0.1:9339";

    info!("starting server...");
    resources::init();
    info!("resources initialized");

    let listener = TcpListener::bind(TCP_ADDR).unwrap();
    while let Ok((stream, addr)) = listener.accept() {
        info!("new connection from {addr}");
        let fd = stream.into_raw_fd();

        thread::spawn(move || receive_loop(fd, addr));
    }
}

fn receive_loop(fd: i32, addr: SocketAddr) {
    static MESSAGE_FACTORY: LazyLock<LogicScrollMessageFactory> =
        LazyLock::new(|| LogicScrollMessageFactory::new(false));

    let mut messaging = Messaging::new(80);

    let encrypter = RC4Encrypter::new(LogicScrollMessageFactory::RC4_KEY, "nonce");
    let decrypter = RC4Encrypter::new(LogicScrollMessageFactory::RC4_KEY, "nonce");

    messaging.get_connection_mut().fd = fd;
    messaging.get_connection_mut().is_connected = true;
    messaging.set_encrypters(encrypter, decrypter);
    messaging.set_message_factory(&MESSAGE_FACTORY);

    let mut session = PlayerSession::new(messaging);
    session.work_until_disconnect();

    info!("client from {addr} disconnected");
}

fn init_tracing() {
    use tracing::level_filters::LevelFilter;
    use tracing_subscriber::{layer::SubscriberExt, util::SubscriberInitExt};

    tracing_subscriber::registry()
        .with(tracing_subscriber::fmt::layer())
        .with(tracing_android::layer("SCROLL-SERVER").unwrap())
        .with(
            tracing_subscriber::EnvFilter::builder()
                .with_default_directive(LevelFilter::INFO.into())
                .from_env_lossy(),
        )
        .init();
}
