use tracing::{error, info};

use crate::{
    network::message::{
        EndClientTurnMessage, GoHomeMessage, OutOfSyncMessage,
        StartMissionMessage,
    },
    session::{PlayerSaveData, PlayerSession},
};

pub fn on_go_home_message(session: &mut PlayerSession, _message: &GoHomeMessage) {
    session.go_home();
}

pub fn on_end_client_turn_message(session: &mut PlayerSession, message: &EndClientTurnMessage) {
    let Some(logic_game_mode) = session.logic_game_mode.as_mut() else {
        error!("EndClientTurnMessage received but LogicGameMode is NULL!");
        return;
    };

    info!(
        "EndClientTurnMessage received: tick: {}, checksum: {}, command count: {}",
        message.tick, message.checksum, message.commands.count
    );

    while logic_game_mode.time.tick < message.tick {
        for command in message.commands.as_slice() {
            if command.execute_tick == logic_game_mode.time.tick {
                info!(
                    "received command of type: {}",
                    (command.vtable.get_command_type)()
                );
                logic_game_mode.logic_command_manager.add_command(command);
            }
        }

        logic_game_mode.update_one_tick();
    }

    let checksum = logic_game_mode.calculate_checksum();
    if checksum != message.checksum {
        error!(
            "client and server are out of sync! tick: {}, server checksum: {}, client checksum: {}",
            logic_game_mode.time.tick, checksum, message.checksum
        );

        let out_of_sync_message = OutOfSyncMessage::new();
        out_of_sync_message.server_checksum = checksum;
        out_of_sync_message.client_checksum = message.checksum;
        out_of_sync_message.tick = logic_game_mode.time.tick;
        session.messaging.send(out_of_sync_message);
    }

    if let Some(logic_client_home) = logic_game_mode.home.get() {
        // Save data if we're in home state

        session.player_save_data = Some(PlayerSaveData::new(
            logic_client_home,
            logic_game_mode.player_avatar.get().unwrap(),
        ));
    }
}

pub fn on_start_mission_message(session: &mut PlayerSession, message: &StartMissionMessage) {
    info!(
        "StartMissionMessage received: npc: {}, location: {}",
        message.npc.parent.global_id, message.npc.location.parent.global_id
    );

    session.start_mission(message.npc);
}
