use std::mem::MaybeUninit;

use crate::ffi_util::import;

#[repr(C)]
pub struct ChecksumEncoder {
    pub vtable: usize,
    pub is_checksum_enabled: bool,
    pub checksum: i32,
    pub checksum_snapshot: i32,
}

#[repr(C)]
pub struct ByteStream {
    pub parent: ChecksumEncoder,
    pub offset: i32,
    pub length: i32,
    pub bit_idx: i32,
    pub data: *const u8,
    pub capacity: i32,
}

impl ByteStream {
    pub fn new(initial_capacity: usize) -> Self {
        import!(byte_stream_ctor(ptr: *mut ByteStream, initial_capacity: i32) -> () = 0x17B098 + 1);
        unsafe {
            let mut instance = MaybeUninit::zeroed().assume_init();
            byte_stream_ctor(&mut instance, initial_capacity as i32);
            instance
        }
    }

    pub fn write_byte(&mut self, value: u8) {
        import!(byte_stream_write_byte(ptr: *mut ByteStream, value: u8) -> () = 0x17B7EC + 1);
        byte_stream_write_byte(self, value);
    }

    pub fn reset_offset(&mut self) {
        self.offset = 0;
        self.bit_idx = 0;
    }

    pub fn get_length(&self) -> usize {
        std::cmp::max(self.offset, self.length) as usize
    }

    pub fn get_data(&self) -> &[u8] {
        unsafe { std::slice::from_raw_parts(self.data, self.get_length()) }
    }

    pub fn destruct(mut self) {
        import!(byte_stream_destruct(ptr: *mut ByteStream) -> () = 0x17BC20 + 1);
        byte_stream_destruct(&mut self);
    }
}

impl<T> From<T> for ByteStream
where
    T: AsRef<[u8]>,
{
    fn from(value: T) -> Self {
        import!(byte_stream_ctor(ptr: *mut ByteStream, data: *const u8, length: i32) -> () = 0x17B0E0 + 1);

        unsafe {
            let mut instance = MaybeUninit::zeroed().assume_init();
            byte_stream_ctor(
                &mut instance,
                value.as_ref().as_ptr(),
                value.as_ref().len() as i32,
            );

            instance
        }
    }
}
